"""
These tasks serves to 1) illustrate what types of tasks *can* be done with Norse
and 2) how to use Norse for specific tasks. Please note that some tasks require
additional dependencies, like OpenAI gym for the cartpole task, which is not included in
vanilla Norse.

Please refer to the page on :doc:`/tasks` for thorough explanations on how to execute and run the
experiments/tasks.
"""
